#![no_main]
#![no_std]

extern crate cortex_m;
#[macro_use(entry, exception)]
extern crate cortex_m_rt as rt;
#[macro_use(block)]
extern crate nb;
extern crate panic_semihosting;

extern crate stm32f4xx_hal as hal;
// #[macro_use(block)]
// extern crate nb;

use core::borrow::BorrowMut;
use core::cell::RefCell;

use core2::io::{Cursor, Write};
use cortex_m::interrupt::{free, Mutex};
use cortex_m::peripheral::NVIC;
use hal::gpio::{self, Alternate, Pin, Pull, PushPull};
use hal::pac::{self, SPI2};
use hal::spi::{Event, Mode, Phase, Polarity, Spi, SpiSlave};
use rf4463::config::RADIO_CONFIG_500_2;
use rf4463::Rf4463;
use ringbuf::{Rb, StaticRb};

use crate::hal::{interrupt, prelude::*, serial::Serial};
use crate::rt::ExceptionFrame;

enum SlaveCmd {
    BufferStatus,
    SendPacket,
}

impl SlaveCmd {
    pub fn from_u8(i: u8) -> Option<Self> {
        match i {
            0 => Some(Self::BufferStatus),
            1 => Some(Self::SendPacket),
            _ => None,
        }
    }
}

enum SlaveState {
    Idle,
    BufferStatus,
    RecvPacket(usize),
}

static SPI_SLAVE: Mutex<RefCell<Option<SpiSlave<SPI2, false, u8>>>> =
    Mutex::new(RefCell::new(None));

static SLAVE_STATE: Mutex<RefCell<SlaveState>> = Mutex::new(RefCell::new(SlaveState::Idle));

static BUF: Mutex<RefCell<Option<StaticRb<u8, 16384>>>> = Mutex::new(RefCell::new(None));

const MODE: Mode = Mode {
    polarity: Polarity::IdleLow,
    phase: Phase::CaptureOnFirstTransition,
};

#[entry]
fn main() -> ! {
    let cp = cortex_m::Peripherals::take().unwrap();
    let dp = pac::Peripherals::take().unwrap();

    let mut rcc = dp.RCC.constrain();

    let mut gpioa = dp.GPIOA.split();
    let mut gpiob = dp.GPIOB.split();

    // setup clocks
    let clocks = rcc.cfgr.use_hse(25.MHz()).sysclk(48.MHz()).freeze();

    // setup 4463 spi
    let mosi = gpioa.pa7.into_alternate();
    let miso = gpioa.pa6.into_alternate();
    let sclk = gpioa.pa5.into_alternate();

    let sdn = gpioa.pa1.into_push_pull_output();

    let cs = gpioa.pa4.into_push_pull_output();

    let spi = Spi::new(dp.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

    let delay = dp.TIM2.delay_ms(&clocks);
    let mut radio = Rf4463::new(spi, sdn, cs, delay, &RADIO_CONFIG_500_2).unwrap();

    // setup Pi SPI (we're the slave)
    let sck2 = gpiob.pb13.internal_resistor(Pull::Up);
    let miso2 = gpiob.pb14.internal_resistor(Pull::Down);
    let mosi2 = gpiob.pb15.internal_resistor(Pull::Down);
    let nss = gpiob.pb12.internal_resistor(Pull::Up).into(); // TODO

    let mut spi2 = dp.SPI2.spi_slave((sck2, miso2, mosi2, Some(nss)), MODE);

    // fill up the tx buffer
    while spi2.is_tx_empty() {
        spi2.write(&[0x00]);
    }

    spi2.listen(Event::Rxne);

    free(|cs| {
        BUF.borrow(cs).replace(Some(StaticRb::default()));
        SPI_SLAVE.borrow(cs).replace(Some(spi2));
    });

    // enable spi2 interrupt
    unsafe {
        NVIC::unmask(interrupt::SPI2);
    }

    const LEN: usize = 256 + 2 + 65;
    let mut data = [0; LEN];
    let mut i = 0;

    loop {
        if i == LEN {
            radio.tx(&data).unwrap();
            i = 0;
            continue;
        }

        free(|cs| {
            let mut buf = BUF.borrow(cs).borrow_mut();
            let buf = buf.as_mut().unwrap();

            if let Some(b) = buf.pop() {
                data[i] = b;

                i += 1;

                if buf.free_len() == 16000 {
                    //block!(USART.borrow(cs).borrow_mut().as_mut().unwrap().write(b'E')).unwrap();
                }
            }
        });
    }
}

#[exception]
unsafe fn HardFault(ef: &ExceptionFrame) -> ! {
    panic!("{:#?}", ef);
}

#[interrupt]
fn SPI2() {
    free(|cs| {
        let mut spi = SPI_SLAVE.borrow(cs).borrow_mut();
        let spi = spi.as_mut().unwrap();

        let x = [0];
        spi.read(&mut x).unwrap();
        spi.write(&[0x00]).unwrap();

        let mut buf = BUF.borrow(cs).borrow_mut();
        let buf = buf.as_mut().unwrap();

        buf.push(x).unwrap();

        if buf.free_len() == 400 {
            block!(us.write(b'F')).unwrap();
        }
    });
}
